/* -*-c-*-
 *
 * General block cipher utilities
 *
 * (c) 2000 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include <assert.h>

#include "gcipher.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @keysz@ --- *
 *
 * Arguments:	@size_t sz@ = a proposed key size, or zero
 *		@const octet *ksz@ = pointer to key size table
 *
 * Returns:	See below.
 *
 * Use:		Returns a sensible key size.  If @sz@ is nonzero, it is
 *		interpreted as an amount (in bytes) of key material which the
 *		caller has available, and the return value is either the
 *		largest allowable key size less than or equal to the caller's
 *		size, or zero if there is no valid key length small enough.
 *		If @sz@ is zero, the function returns a `recommended' key
 *		size.
 */

size_t keysz(size_t sz, const octet *ksz)
{
  unsigned op = ksz[0]&KSZ_OPMASK;
  unsigned wd = (ksz[0]&KSZ_16BIT) ? 2 : 1;
  unsigned t, u, v;

  ksz++;
#define ARG(i) (wd == 1 ? ksz[i] : LOAD16(ksz + 2*i))
  if (sz == 0)
    return (ARG(0));
  else switch (op) {
    case KSZ_ANY: return (sz);
    case KSZ_RANGE:
      t = ARG(1); u = ARG(2); v = ARG(3);
      if (v) sz -= sz%v;
      if (sz < t) return (0);
      if (u && sz > u) return (u);
      return (sz);
    case KSZ_SET:
      u = 0;
      for (;;) {
	t = ARG(0); ksz += wd; if (!t) break;
	if (sz >= t && u < t) u = t;
      }
      return (u);
  }
#undef ARG

  assert(((void)"bad key size table", 0));
  return (0);
}

/*----- That's all, folks -------------------------------------------------*/
