/* -*-c-*-
 *
 * Implementation of the SHA-256 hash function
 *
 * (c) 2000 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include <mLib/bits.h>

#include "ghash.h"
#include "ghash-def.h"
#include "hash.h"
#include "sha256.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @sha256_compress@, @sha224_compress@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context block
 *		@const void *sbuf@ = pointer to buffer of appropriate size
 *
 * Returns:	---
 *
 * Use:		SHA-256 compression function.
 */

void sha256_compress(sha256_ctx *ctx, const void *sbuf)
{
  uint32 a, b, c, d, e, f, g, h;
  uint32 m[16];
  const octet *p;
  int i;

  a  = ctx->a; b  = ctx->b; c  = ctx->c; d  = ctx->d;
  e  = ctx->e; f  = ctx->f; g  = ctx->g; h  = ctx->h;
  for (p = sbuf, i = 0; i < 16; i++, p += 4) m[i] = LOAD32(p);

  /* --- Definitions for round functions --- */

#define CH(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define MAJ(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))

#define S0(x) (ROR32((x),  2) ^ ROR32((x), 13) ^ ROR32((x), 22))
#define S1(x) (ROR32((x),  6) ^ ROR32((x), 11) ^ ROR32((x), 25))
#define s0(x) (ROR32((x),  7) ^ ROR32((x), 18) ^ LSR32((x),  3))
#define s1(x) (ROR32((x), 17) ^ ROR32((x), 19) ^ LSR32((x), 10))

#define T(a, b, c, d, e, f, g, h, i, k) do {				\
  uint32 t1 = h + S1(e) + CH(e, f, g) + k + m[i];			\
  uint32 t2 = S0(a) + MAJ(a, b, c);					\
  d += t1; h = t1 + t2;							\
} while (0)

#define M(i, i2, i7, i15)						\
  do { m[i] += s1(m[i2]) + m[i7] + s0(m[i15]); } while (0)

  /* --- The main compression function --- */

  T(a, b, c, d, e, f, g, h,  0, 0x428a2f98);	M( 0, 14,  9,  1);
  T(h, a, b, c, d, e, f, g,  1, 0x71374491);	M( 1, 15, 10,  2);
  T(g, h, a, b, c, d, e, f,  2, 0xb5c0fbcf);	M( 2,  0, 11,  3);
  T(f, g, h, a, b, c, d, e,  3, 0xe9b5dba5);	M( 3,  1, 12,  4);
  T(e, f, g, h, a, b, c, d,  4, 0x3956c25b);	M( 4,  2, 13,  5);
  T(d, e, f, g, h, a, b, c,  5, 0x59f111f1);	M( 5,  3, 14,  6);
  T(c, d, e, f, g, h, a, b,  6, 0x923f82a4);	M( 6,  4, 15,  7);
  T(b, c, d, e, f, g, h, a,  7, 0xab1c5ed5);	M( 7,  5,  0,  8);
  T(a, b, c, d, e, f, g, h,  8, 0xd807aa98);	M( 8,  6,  1,  9);
  T(h, a, b, c, d, e, f, g,  9, 0x12835b01);	M( 9,  7,  2, 10);
  T(g, h, a, b, c, d, e, f, 10, 0x243185be);	M(10,  8,  3, 11);
  T(f, g, h, a, b, c, d, e, 11, 0x550c7dc3);	M(11,  9,  4, 12);
  T(e, f, g, h, a, b, c, d, 12, 0x72be5d74);	M(12, 10,  5, 13);
  T(d, e, f, g, h, a, b, c, 13, 0x80deb1fe);	M(13, 11,  6, 14);
  T(c, d, e, f, g, h, a, b, 14, 0x9bdc06a7);	M(14, 12,  7, 15);
  T(b, c, d, e, f, g, h, a, 15, 0xc19bf174);	M(15, 13,  8,  0);
  T(a, b, c, d, e, f, g, h,  0, 0xe49b69c1);	M( 0, 14,  9,  1);
  T(h, a, b, c, d, e, f, g,  1, 0xefbe4786);	M( 1, 15, 10,  2);
  T(g, h, a, b, c, d, e, f,  2, 0x0fc19dc6);	M( 2,  0, 11,  3);
  T(f, g, h, a, b, c, d, e,  3, 0x240ca1cc);	M( 3,  1, 12,  4);
  T(e, f, g, h, a, b, c, d,  4, 0x2de92c6f);	M( 4,  2, 13,  5);
  T(d, e, f, g, h, a, b, c,  5, 0x4a7484aa);	M( 5,  3, 14,  6);
  T(c, d, e, f, g, h, a, b,  6, 0x5cb0a9dc);	M( 6,  4, 15,  7);
  T(b, c, d, e, f, g, h, a,  7, 0x76f988da);	M( 7,  5,  0,  8);
  T(a, b, c, d, e, f, g, h,  8, 0x983e5152);	M( 8,  6,  1,  9);
  T(h, a, b, c, d, e, f, g,  9, 0xa831c66d);	M( 9,  7,  2, 10);
  T(g, h, a, b, c, d, e, f, 10, 0xb00327c8);	M(10,  8,  3, 11);
  T(f, g, h, a, b, c, d, e, 11, 0xbf597fc7);	M(11,  9,  4, 12);
  T(e, f, g, h, a, b, c, d, 12, 0xc6e00bf3);	M(12, 10,  5, 13);
  T(d, e, f, g, h, a, b, c, 13, 0xd5a79147);	M(13, 11,  6, 14);
  T(c, d, e, f, g, h, a, b, 14, 0x06ca6351);	M(14, 12,  7, 15);
  T(b, c, d, e, f, g, h, a, 15, 0x14292967);	M(15, 13,  8,  0);
  T(a, b, c, d, e, f, g, h,  0, 0x27b70a85);	M( 0, 14,  9,  1);
  T(h, a, b, c, d, e, f, g,  1, 0x2e1b2138);	M( 1, 15, 10,  2);
  T(g, h, a, b, c, d, e, f,  2, 0x4d2c6dfc);	M( 2,  0, 11,  3);
  T(f, g, h, a, b, c, d, e,  3, 0x53380d13);	M( 3,  1, 12,  4);
  T(e, f, g, h, a, b, c, d,  4, 0x650a7354);	M( 4,  2, 13,  5);
  T(d, e, f, g, h, a, b, c,  5, 0x766a0abb);	M( 5,  3, 14,  6);
  T(c, d, e, f, g, h, a, b,  6, 0x81c2c92e);	M( 6,  4, 15,  7);
  T(b, c, d, e, f, g, h, a,  7, 0x92722c85);	M( 7,  5,  0,  8);
  T(a, b, c, d, e, f, g, h,  8, 0xa2bfe8a1);	M( 8,  6,  1,  9);
  T(h, a, b, c, d, e, f, g,  9, 0xa81a664b);	M( 9,  7,  2, 10);
  T(g, h, a, b, c, d, e, f, 10, 0xc24b8b70);	M(10,  8,  3, 11);
  T(f, g, h, a, b, c, d, e, 11, 0xc76c51a3);	M(11,  9,  4, 12);
  T(e, f, g, h, a, b, c, d, 12, 0xd192e819);	M(12, 10,  5, 13);
  T(d, e, f, g, h, a, b, c, 13, 0xd6990624);	M(13, 11,  6, 14);
  T(c, d, e, f, g, h, a, b, 14, 0xf40e3585);	M(14, 12,  7, 15);
  T(b, c, d, e, f, g, h, a, 15, 0x106aa070);	M(15, 13,  8,  0);
  T(a, b, c, d, e, f, g, h,  0, 0x19a4c116);
  T(h, a, b, c, d, e, f, g,  1, 0x1e376c08);
  T(g, h, a, b, c, d, e, f,  2, 0x2748774c);
  T(f, g, h, a, b, c, d, e,  3, 0x34b0bcb5);
  T(e, f, g, h, a, b, c, d,  4, 0x391c0cb3);
  T(d, e, f, g, h, a, b, c,  5, 0x4ed8aa4a);
  T(c, d, e, f, g, h, a, b,  6, 0x5b9cca4f);
  T(b, c, d, e, f, g, h, a,  7, 0x682e6ff3);
  T(a, b, c, d, e, f, g, h,  8, 0x748f82ee);
  T(h, a, b, c, d, e, f, g,  9, 0x78a5636f);
  T(g, h, a, b, c, d, e, f, 10, 0x84c87814);
  T(f, g, h, a, b, c, d, e, 11, 0x8cc70208);
  T(e, f, g, h, a, b, c, d, 12, 0x90befffa);
  T(d, e, f, g, h, a, b, c, 13, 0xa4506ceb);
  T(c, d, e, f, g, h, a, b, 14, 0xbef9a3f7);
  T(b, c, d, e, f, g, h, a, 15, 0xc67178f2);

  /* --- Update the chaining variables --- */

  ctx->a += a; ctx->b += b; ctx->c += c; ctx->d += d;
  ctx->e += e; ctx->f += f; ctx->g += g; ctx->h += h;
}

/* --- @sha256_init@, @sha224_init@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context block to initialize
 *
 * Returns:	---
 *
 * Use:		Initializes a context block ready for hashing.
 */

void sha256_init(sha256_ctx *ctx)
{
  ctx->a = 0x6a09e667;
  ctx->b = 0xbb67ae85;
  ctx->c = 0x3c6ef372;
  ctx->d = 0xa54ff53a;
  ctx->e = 0x510e527f;
  ctx->f = 0x9b05688c;
  ctx->g = 0x1f83d9ab;
  ctx->h = 0x5be0cd19;
  ctx->off = 0;
  ctx->nl = ctx->nh = 0;
}

void sha224_init(sha256_ctx *ctx)
{
  ctx->a = 0xc1059ed8;
  ctx->b = 0x367cd507;
  ctx->c = 0x3070dd17;
  ctx->d = 0xf70e5939;
  ctx->e = 0xffc00b31;
  ctx->f = 0x68581511;
  ctx->g = 0x64f98fa7;
  ctx->h = 0xbefa4fa4;
  ctx->off = 0;
  ctx->nl = ctx->nh = 0;
}

/* --- @sha256_set@, @sha224_set@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = pointer to state buffer
 *		@unsigned long count@ = current count of bytes processed
 *
 * Returns:	---
 *
 * Use:		Initializes a context block from a given state.  This is
 *		useful in cases where the initial hash state is meant to be
 *		secret, e.g., for NMAC and HMAC support.
 */

void sha256_set(sha256_ctx *ctx, const void *buf, unsigned long count)
{
  const octet *p = buf;
  ctx->a = LOAD32(p +  0);
  ctx->b = LOAD32(p +  4);
  ctx->c = LOAD32(p +  8);
  ctx->d = LOAD32(p + 12);
  ctx->e = LOAD32(p + 16);
  ctx->f = LOAD32(p + 20);
  ctx->g = LOAD32(p + 24);
  ctx->h = LOAD32(p + 28);
  ctx->off = 0;
  ctx->nl = U32(count);
  ctx->nh = U32(((count & ~(unsigned long)MASK32) >> 16) >> 16);
}

/* --- @sha256_hash@, @sha224_hash@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context block
 *		@const void *buf@ = buffer of data to hash
 *		@size_t sz@ = size of buffer to hash
 *
 * Returns:	---
 *
 * Use:		Hashes a buffer of data.  The buffer may be of any size and
 *		alignment.
 */

void sha256_hash(sha256_ctx *ctx, const void *buf, size_t sz)
{
  HASH_BUFFER(SHA256, sha256, ctx, buf, sz);
}

/* --- @sha256_done, @sha224_done@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context block
 *		@void *hash@ = pointer to output buffer
 *
 * Returns:	---
 *
 * Use:		Returns the hash of the data read so far.
 */

static void final(sha256_ctx *ctx)
{
  HASH_PAD(SHA256, sha256, ctx, 0x80, 0, 8);
  STORE32(ctx->buf + SHA256_BUFSZ - 8, (ctx->nl >> 29) | (ctx->nh << 3));
  STORE32(ctx->buf + SHA256_BUFSZ - 4, ctx->nl << 3);
  sha256_compress(ctx, ctx->buf);
}

void sha256_done(sha256_ctx *ctx, void *hash)
{
  octet *p = hash;
  final(ctx);
  STORE32(p +  0, ctx->a);
  STORE32(p +  4, ctx->b);
  STORE32(p +  8, ctx->c);
  STORE32(p + 12, ctx->d);
  STORE32(p + 16, ctx->e);
  STORE32(p + 20, ctx->f);
  STORE32(p + 24, ctx->g);
  STORE32(p + 28, ctx->h);
}

void sha224_done(sha224_ctx *ctx, void *hash)
{
  octet *p = hash;
  final(ctx);
  STORE32(p +  0, ctx->a);
  STORE32(p +  4, ctx->b);
  STORE32(p +  8, ctx->c);
  STORE32(p + 12, ctx->d);
  STORE32(p + 16, ctx->e);
  STORE32(p + 20, ctx->f);
  STORE32(p + 24, ctx->g);
}

/* --- @sha256_state@, @sha224_state@ --- *
 *
 * Arguments:	@sha256_ctx *ctx@ = pointer to context
 *		@void *state@ = pointer to buffer for current state
 *
 * Returns:	Number of bytes written to the hash function so far.
 *
 * Use:		Returns the current state of the hash function such that
 *		it can be passed to @sha256_set@.
 */

unsigned long sha256_state(sha256_ctx *ctx, void *state)
{
  octet *p = state;
  STORE32(p +  0, ctx->a);
  STORE32(p +  4, ctx->b);
  STORE32(p +  8, ctx->c);
  STORE32(p + 12, ctx->d);
  STORE32(p + 16, ctx->e);
  STORE32(p + 20, ctx->f);
  STORE32(p + 24, ctx->g);
  STORE32(p + 28, ctx->h);
  return (ctx->nl | ((ctx->nh << 16) << 16));
}

/* --- Generic interface --- */

#define HASHES(_)							\
  _(SHA224, sha224)							\
  _(SHA256, sha256)

HASHES(GHASH_DEF)

/*----- Test rig ----------------------------------------------------------*/

#ifdef TEST_RIG

#include <mLib/testrig.h>

HASHES(HASH_VERIFY)

static const test_chunk defs[] = {
  HASHES(HASH_TESTDEFS)
  { 0, 0, { 0 } }
};

int main(int argc, char *argv[])
{
  test_run(argc, argv, defs, SRCDIR "/t/sha256");
  return (0);
}

#endif

/*----- That's all, folks -------------------------------------------------*/
